"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import DictStrAny
from encommon.types import NCNone

from .child import RobieChild
from ..addons import RobieQueue
from ..models import RobieModels
from ...utils import ClientChannels

if TYPE_CHECKING:
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..params import RobieClientParams
    from ..threads import RobieClientThread



class RobieClient(RobieChild):
    """
    Establish and maintain connection with the chat service.
    """

    __thread: Optional['RobieClientThread'] = None


    @property
    def family(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kind(
        self,
    ) -> Literal['client']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'client'


    @property
    def channels(
        self,
    ) -> ClientChannels:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @classmethod
    def schema(
        cls,
    ) -> Type['RobieClientParams']:
        """
        Return the configuration parameters relevant for class.

        :returns: Configuration parameters relevant for class.
        """

        raise NotImplementedError


    @property
    def params(
        self,
    ) -> 'RobieClientParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        model = (
            RobieModels
            .client())

        params = super().params

        assert isinstance(
            params, model)

        return params


    @property
    def thread(
        self,
    ) -> Optional['RobieClientThread']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        from ..threads import (
            RobieClientThread)

        thread = self.__thread

        if thread is None:
            return NCNone

        assert isinstance(
            thread,
            RobieClientThread)

        return thread


    @thread.setter
    def thread(
        self,
        value: 'RobieClientThread',
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        from ..threads import (
            RobieClientThread)

        assert isinstance(
            value,
            RobieClientThread)

        self.__thread = value


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        dumped = super().dumped

        family = self.family

        return dumped | {
            'family': family}


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        .. note::
           Deviates from enhomie in children have operations,
           and are more isolated from internal core routines.
        """

        raise NotImplementedError


    def get_message(
        self,
    ) -> 'RobieMessage':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param mqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def get_command(
        self,
    ) -> 'RobieCommand':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param cqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from pydantic_ai import Agent
    from pydantic_ai.settings import ModelSettings
    from pydantic_ai.models.anthropic import AnthropicModel
    from pydantic_ai.models.openai import OpenAIModel
    from .common import AinswerDepends
    from .plugin import AinswerPlugin



class AinswerModels:
    """
    Return the class object that was imported within method.

    :param plugin: Plugin class instance for Chatting Robie.
    """

    __plugin: 'AinswerPlugin'


    def __init__(
        self,
        plugin: 'AinswerPlugin',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__plugin = plugin


    @classmethod
    def agent(
        cls,
    ) -> Type['Agent[AinswerDepends, str]']:
        """
        Return the class object that was imported within method.
        """

        from pydantic_ai import Agent

        return Agent


    @classmethod
    def settings(
        cls,
    ) -> Type['ModelSettings']:
        """
        Return the class object that was imported within method.
        """

        from pydantic_ai.settings import ModelSettings

        return ModelSettings


    @property
    def anthropic(
        self,
    ) -> 'AnthropicModel':
        """
        Return the class object that was imported within method.
        """

        from pydantic_ai.models.anthropic import AnthropicModel
        from pydantic_ai.providers.anthropic import AnthropicProvider

        plugin = self.__plugin

        params = plugin.params
        ainswer = params.ainswer
        secret = ainswer.secret

        provider = (
            AnthropicProvider(
                api_key=secret))

        return AnthropicModel(
            ainswer.model,
            provider=provider)


    @property
    def openai(
        self,
    ) -> 'OpenAIModel':
        """
        Return the class object that was imported within method.
        """

        from pydantic_ai.models.openai import OpenAIModel
        from pydantic_ai.providers.openai import OpenAIProvider

        plugin = self.__plugin

        params = plugin.params
        ainswer = params.ainswer
        secret = ainswer.secret

        provider = (
            OpenAIProvider(
                api_key=secret))

        return OpenAIModel(
            ainswer.model,
            provider=provider)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import BaseModel

from pydantic import Field

if TYPE_CHECKING:
    from .plugin import AinswerPlugin
    from ...robie.models import RobieMessage



AinswerIgnored = 'no_response'

AinswerTool = Callable[..., Any]



@dataclass
class AinswerDepends:
    """
    Dependencies related to operation with PydanticAI tools.
    """

    plugin: 'AinswerPlugin'
    mitem: Optional['RobieMessage']



class AinswerResponse(BaseModel, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Simple and concise format.')]



class AinswerResponseDSC(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for Discord.',
              max_length=1900)]



class AinswerResponseIRC(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for IRCv2.',
              max_length=350)]



class AinswerResponseMTM(AinswerResponse, extra='forbid'):
    """
    Contains the response for question for the client type.
    """

    text: Annotated[
        str,
        Field(...,
              description='Format intended for Mattermost.',
              max_length=1900)]



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..plugin import AinswerPlugin

if TYPE_CHECKING:
    from ....robie import RobieService



def test_AinswerModels(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    plugin = (
        service.plugins
        .childs['ainswer'])

    assert isinstance(
        plugin, AinswerPlugin)


    models = plugin.models


    attrs = lattrs(models)

    assert attrs == [
        '_AinswerModels__plugin']


    assert inrepr(
        'models.AinswerModels',
        models)

    assert isinstance(
        hash(models), int)

    assert instr(
        'models.AinswerModels',
        models)


    assert models.agent()

    assert models.settings()

    assert models.anthropic

    assert models.openai



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



import asyncio
from typing import Optional

from encommon.types import funcname

from pydantic_ai import RunContext

from .common import AinswerDepends
from .memory import AinswerMemoryMessageField
from .memory import AinswerMemoryRecord
from .memory import AinswerMemoryUniqueField



NOTAPERSON = Exception(
    'User is not a Robie person')



async def ainswer_memory_insert(
    context: RunContext[AinswerDepends],
    message: AinswerMemoryMessageField,
) -> None:
    """
    Insert the memory into the historical person memorables.

    Do not add memories of previous questions or discussion.
    Do not add new memories without confirmation as there is
    limited storage that the person may have for memorables.
    Only add the memory one time after ensuring it does not
    already exist, using the `ainswer_memory_records` tool.
    Always confirm with the person before inserting message.
    Be very cautious and conservative with use of this tool,
    the bot is configured by default to remove older entries
    once the limit for the person memories has been reached!

    .. note::
       This tool will allow for inserting new memories for
       the provided username, also known as a Robie person.
    """

    deps = context.deps
    ainswer = deps.plugin
    mitem = deps.mitem
    memory = ainswer.memory

    inputted = {
        'message': message}

    await asyncio.sleep(0)


    # Because of TestModel
    if mitem is None:
        return None


    _person = mitem.person

    if _person is None:
        return None

    memory.insert(
        person=_person,
        message=message)


    ainswer.printer({
        'Function': funcname(),
        'Inputted': inputted})



async def ainswer_memory_records(
    context: RunContext[AinswerDepends],
) -> Optional[list[AinswerMemoryRecord]]:
    """
    Return the historical records for the person memorables.

    .. note::
       This tool will return historical memorable items for
       the provided username, also known as a Robie person.
    """

    deps = context.deps
    ainswer = deps.plugin
    mitem = deps.mitem


    await asyncio.sleep(0)

    # Because of TestModel
    if mitem is None:
        return None


    _person = mitem.person

    if _person is None:
        return None

    returned = (
        ainswer.memory
        .records(mitem))


    ainswer.printer({
        'Function': funcname(),
        'Returned': returned})

    return returned



async def ainswer_memory_delete(
    context: RunContext[AinswerDepends],
    unique: AinswerMemoryUniqueField,
) -> None:
    """
    Delete the record from the historical person memorables.

    .. note::
       This tool will allow for deleting the memories for
       the provided username, also known as a Robie person.
    """

    deps = context.deps
    ainswer = deps.plugin
    mitem = deps.mitem
    memory = ainswer.memory

    inputted = {
        'unique': unique}


    await asyncio.sleep(0)

    # Because of TestModel
    if mitem is None:
        return None


    _person = mitem.person

    if _person is None:
        return None

    memory.delete(
        _person, unique)


    ainswer.printer({
        'Function': funcname(),
        'Inputted': inputted})
